@@ -42,8 +42,9 @@ class PatternMiner:
     metapatterns : list of dictionaries (optional)
         a metapattern is a dict with
             'name': identifier of the metapattern (optional)
-            'P_columns': columns of dataframe (P part of metapattern)
-            'Q_columns': columns of datafrane (Q part of metapattern)
+            'pattern' / 'expression' : The pattern or expression to be investigated
+            'P_columns': columns of dataframe (P part of metapattern) (optional)
+            'Q_columns': columns of datafrane (Q part of metapattern) (optional)
             'parameters': minimum confidence, patterns with higher confidence are included (optional)
             'encode': encoding definitions of the columns (optional)
 
@@ -71,10 +72,12 @@ class PatternMiner:
     def find(self, *args, **kwargs):
         '''General function to find patterns
         '''
+        logger = logging.getLogger(__name__)
+
         self.__process_parameters(*args, **kwargs)
         assert self.metapatterns is not None, "No patterns defined."
         assert self.df_data is not None, "No dataframe defined."
-        logging.info(' Rows in data: ' + str(self.df_data.shape[0]))
+        logger.info('Rows in data: ' + str(self.df_data.shape[0]))
 
         new_df_patterns = derive_patterns(**kwargs, metapatterns = self.metapatterns, dataframe = self.df_data)
 
@@ -99,14 +102,17 @@ class PatternMiner:
 
         df_data = self.df_data.copy()
         df_results = self.df_results.copy()
+
+        # get Q value that is correct from the pattern
         df_results['correct_value'] = df_results['pattern_def'].apply(lambda x: get_value(x, num=2))
         df_results = df_results.loc[df_results['result_type'] == False]
 
-        # Get the column names
+        # Get the Q column names
         colq = get_value(df_results['pattern_def'].iloc[0], 2, 1)
 
+        # Replace the data with the correct values
         df_data.loc[df_data.index.isin(df_results.index), colq] = df_results['correct_value']
-        return df_data, df_results # changed data and log what changed
+        return df_data, df_results
 
 
     def analyze(self, *args, **kwargs):
@@ -165,34 +171,34 @@ class PatternMiner:
         return current
 
 def get_value(pattern, num = 1, col=3):
-    ''' Derive the P (num=1) or Q (num=2) value from a pattern'''
+    ''' Derive the P (num=1) or Q (num=2) value (col=3) or column name (col=1) from a pattern'''
 
     values = []
 
     item = re.search(r'IF(.*)THEN(.*)', pattern)
 
-    # If not conditional statement, return None
+    # If not conditional statement
     if item == None:
         item2 = re.search(r'(.*)[><=](.*)', pattern)
 
-        if pattern.count('}') > 2: # sum
-            for match in re.finditer(r'{(.*?)}', item2.group(num)):
+        if pattern.count('}') > 2: # sum pattern
+            for match in re.finditer(r'{(.*?)}', item2.group(num)): # get all columns
                 item3 = re.search(r'"(.*)"', match.group(1))
                 values.append(item3.group(1))
 
         else:
-            if '{' in item2.group(2):
-                item3 = re.search(r'{(.*?)}', item2.group(num))
+            if '{' in item2.group(2): # compare column pattern
+                item3 = re.search(r'{(.*?)}', item2.group(num)) # get columns
                 values.append(item3.group(1)[1:-1])
 
-            else:
+            else: # compare value pattern
                 if num == 1:
                     item3 = re.search(r'{(.*?)}', item2.group(1))
                     values.append(item3.group(1)[1:-1])
                 else:
                     return None
     else:
-        if col == 1:
+        if col == 1: # get the col name
             for match in re.finditer(r'{(.*?)}', item.group(num)):
                 item3 = match.group(1)[1:-1]
                 values.append(item3)
@@ -224,26 +230,29 @@ def get_value(pattern, num = 1, col=3):
 def derive_patterns(dataframe   = None,
                     metapatterns = None):
     '''Derive patterns from metapatterns
-       In three flavours:
+       In two flavours:
        - expressions (defined as a string),
-       - conditional rules ('-->'-pattern defined with their columns) and
-       - single rules (defined with their colums)
+       - code patterns, which are given by a dictionary
     '''
+    logger = logging.getLogger(__name__)
+    logger.info('Find started ...')
 
     df_patterns = pd.DataFrame(columns = PATTERNS_COLUMNS)
+    dataframe = dataframe.reset_index()
+
     for metapattern in metapatterns:
         parameters = metapattern.get("parameters", {})
         if "expression" in metapattern.keys():
-            patterns = derive_patterns_from_template_expression(metapattern = metapattern,
+            patterns = derive_patterns_from_template_expression(metapattern = metapattern, # expression
                                                                 dataframe = dataframe)
         else:
-            patterns = derive_patterns_from_code(metapattern = metapattern,
+            patterns = derive_patterns_from_code(metapattern = metapattern, # code pattern
                                               dataframe = dataframe)
-        logging.info(' Total rows in patterns: ' + str(patterns.shape[0]))
+        logger.info('Total rows in patterns: ' + str(patterns.shape[0]))
         if 'min_confidence' in parameters:
             if parameters['min_confidence'] == 'highest': # For when you want the highest confidence in a pattern
                 patterns = get_highest_conf(patterns)
-                logging.info(' Reduction of rows in patterns to: ' + str(patterns.shape[0]))
+                logger.info('Reduction of rows in patterns to: ' + str(patterns.shape[0]))
 
         df_patterns = df_patterns.append(patterns, ignore_index = True)
 
@@ -252,9 +261,15 @@ def derive_patterns(dataframe   = None,
     df_patterns[EXCEPTIONS] = df_patterns[EXCEPTIONS].astype(np.int64)
     df_patterns.index.name = 'index'
 
+    logger.info('Find ended ...')
+
     return PatternDataFrame(df_patterns)
 
 def get_highest_conf(df_patterns):
+    """
+    This code takes the highest confidence patterns based on the P expression. This makes it easier to find new conditional patterns,
+    because we do not have to fill in a specific value for the confidence.
+    """
 
     df_patterns['P_val'] = df_patterns['pattern_def'].apply(lambda x: get_value(x, num=1))
 
@@ -269,13 +284,13 @@ def get_highest_conf(df_patterns):
 def derive_patterns_from_template_expression(metapattern = None,
                                              dataframe = None):
     """
-    Here we can add the constructions of expressions from an expression with wildcards
+    Takes the expression and splits it into a condition or quantitative expression
     """
     expression = metapattern.get("expression", "")
     parameters = metapattern.get("parameters", {})
     encodings = metapattern.get("encode", {})
-    solvency = parameters.get('solvency', False)
-    if re.search(r'IF(.*)THEN(.*)', expression):
+
+    if re.search(r'IF(.*)THEN(.*)', expression): # conditional statement
         new_list = derive_patterns_from_expression(expression, metapattern, dataframe)
     else:
         new_list = derive_quantitative_pattern_expression(expression, metapattern, dataframe)
@@ -284,12 +299,20 @@ def derive_patterns_from_template_expression(metapattern = None,
 
 
 def derive_quantitative_pattern_expression(expression, metapattern, dataframe):
-    logging.info(' Calling function derive_quantitative_pattern_expression')
+    """
+    This function takes the quantitative expression and finds all necessary values to put it in the right quantitative code. We need
+    to do this, so that we can use numpy. Numpy runs a lot faster than pandas.
+    """
+    logger = logging.getLogger(__name__)
+
+    logger.info('Calling function derive_quantitative_pattern_expression')
 
     parameters = metapattern.get("parameters", {})
     name = metapattern.get('name', "No name")
 
     confidence, support = get_parameters(parameters)
+
+    # only use numerical columns for quantitative expressions
     numerical_columns = [dataframe.columns[c] for c in range(len(dataframe.columns))
                             if ((dataframe.dtypes[c] == 'float64') or (dataframe.dtypes[c] == 'int64'))]
     dataframe = dataframe[numerical_columns]
@@ -333,7 +356,7 @@ def derive_quantitative_pattern_expression(expression, metapattern, dataframe):
             pattern += '='
             item3 = item3[1:]
 
-        if '{' in item3:
+        if '{' in item3: # column comparing pattern
             P_columns = get_possible_columns(item2.group(1).count('.*'), item2.group(1),dataframe,True)
             Q_columns = get_possible_columns(item3.count('.*'), item3, dataframe,True)
             Q_columns = [dataframe.columns.get_loc(c) for c in Q_columns if c in numerical_columns]
@@ -349,7 +372,7 @@ def derive_quantitative_pattern_expression(expression, metapattern, dataframe):
             for pat in compares:
                 patterns.extend(pat)
 
-        else:
+        else: # column value
             columns = get_possible_columns(item2.group(1).count('.*'), item2.group(1),dataframe,True)
             value = float(item3)
             columns = [dataframe.columns.get_loc(c) for c in columns if c in numerical_columns]
@@ -365,6 +388,11 @@ def derive_quantitative_pattern_expression(expression, metapattern, dataframe):
     return patterns
 
 def get_possible_columns(amount, expression, dataframe, quant=False):
+    """
+    Get the possible columns for the conitional expression using regex
+    """
+
+    # no regex columns, then jut return it
     if amount == 0:
         if quant:
             return [expression.strip()[1:-1]]
@@ -403,11 +431,17 @@ def get_possible_columns(amount, expression, dataframe, quant=False):
     for columns in possibilities:
         possible_expression = expression
         for column in columns: # replace with the possible column value
-            possible_expression = possible_expression.replace(".*", '"' + column + '"', 1) # replace with column
+            possible_expression = possible_expression.replace("{.*}", '{"' + column + '"}', 1) # replace with column
         possible_expressions.append(possible_expression)
     return possible_expressions
 
 def get_possible_values(amount, possible_expressions, dataframe):
+
+    """
+    Get the possible values for the conitional expression
+
+    """
+
     if amount < 1: # no values to be found
         return possible_expressions
     else:
@@ -419,20 +453,39 @@ def get_possible_values(amount, possible_expressions, dataframe):
                 value_col = value_col[2:-2] # strip { and }
                 all_columns.append(value_col)
 
-            # print(all_columns)
             all_columns_v = dataframe[all_columns].drop_duplicates().values
-            # print(all_columns_v)
-            for columns_v in all_columns_v: # Make all combinations without duplicates  of values
+
+            items =  re.findall(r'\[(.*?@)\]', possible_expression) # Find the columns
+            del_rows = []
+            for i in range(len(items)):
+                if len(items[i]) > 1: # only when we have a string
+                    item = items[i][:-1]
+                    for j in range(len(all_columns_v)): # check if each column can be used
+                        if re.search(item, all_columns_v[j][i]) is not None:
+                            if re.search(item, all_columns_v[j][i])[0] != all_columns_v[j][i]:
+                                del_rows.append(j)
+                        else:
+                                del_rows.append(j)
+
+                    possible_expression = possible_expression.replace(item, '', 1) # Replace it so that it goes well later
+
+            all_columns_v = np.delete(all_columns_v, del_rows,0) # del rows
+
+            # Replace @ with the possibilities
+            for columns_v in all_columns_v:
                 possible_expression_v = possible_expression
                 for column_v in columns_v:
                     if isinstance(column_v, str):
-                        possible_expression_v = possible_expression_v.replace('"@"', '"'+ column_v +'"', 1) # replace adn add ""
+                        possible_expression_v = possible_expression_v.replace('[@]', '"'+ column_v +'"', 1) # replace adn add ""
                     else:
-                        possible_expression_v = possible_expression_v.replace('"@"', str(column_v), 1) # replace with str
+                        possible_expression_v = possible_expression_v.replace('[@]', str(column_v), 1) # replace with str
                 expressions.append(possible_expression_v)
         return expressions
 
 def add_qoutation(possible_expressions):
+    """
+    Add quotation at the right places
+    """
     new_expressions = []
     datapoints = []
     for expression in possible_expressions:
@@ -449,8 +502,11 @@ def derive_patterns_from_expression(expression = "",
                                     metapattern = None,
                                     dataframe = None):
     """
+    Derive pattern from expression string. This will be a condit expression
     """
-    logging.info(' Calling function derive_patterns_from_expression')
+    logger = logging.getLogger(__name__)
+
+    logger.info('Calling function derive_patterns_from_expression')
 
     parameters = metapattern.get("parameters", {})
     name = metapattern.get('name', "No name")
@@ -460,7 +516,6 @@ def derive_patterns_from_expression(expression = "",
     if confidence == 'highest':
         confidence = 0
 
-    solvency = parameters.get('solvency', False)
     patterns = list()
 
     amount = expression.count('.*}') #Amount of columns to be found
@@ -479,45 +534,40 @@ def derive_patterns_from_expression(expression = "",
     else:
         dataframe2 = dataframe
 
+    # Check for the possible columns and values
     possible_expressions = get_possible_columns(amount, expression, dataframe2)
     possible_expressions = add_qoutation(possible_expressions)
     possible_expressions = get_possible_values(amount_v, possible_expressions, dataframe2)
+
+    # Log if it is a high value
     if len(possible_expressions) > 40:
-        logging.warning(' Amount of possibilities is high! Namely, ' + str(len(possible_expressions)))
+        logger.warning(' Amount of possibilities is high! Namely, ' + str(len(possible_expressions)))
     else:
-        logging.info(' Amount of possibilities: ' + str(len(possible_expressions)))
+        logger.info(' Amount of possibilities: ' + str(len(possible_expressions)))
 
     for possible_expression in possible_expressions:
-        # print(possible_expression)
         pandas_expressions = to_pandas_expressions(possible_expression, encode, parameters, dataframe)
-        # print(pandas_expressions)
         try: # Some give error so we use try
             n_co = len(eval(pandas_expressions[0], encodings, {'df': dataframe, 'MAX': np.maximum, 'MIN': np.minimum, 'SUM': np.sum}).index)
             n_ex = len(eval(pandas_expressions[1], encodings, {'df': dataframe, 'MAX': np.maximum, 'MIN': np.minimum, 'SUM': np.sum}).index)
             conf = np.round(n_co / (n_co + n_ex + 1e-11), 4)
-            # print(n_co,n_ex)
             if ((conf >= confidence) and (n_co >= support)):
                 xbrl_expressions = to_xbrl_expressions(possible_expression, encode, parameters)
                 patterns.extend([[[name, 0], possible_expression, [n_co, n_ex, conf]] + pandas_expressions + xbrl_expressions + ['']])
+                logger.info("Pattern " + name + " correctly parsed (#co=" + str(co) + ", #ex=" + str(ex) + ")")
         except TypeError as e:
-            logging.Error('Error in trying to process pandas expression: ' +  str(e))
-            if solvency:
-                patterns.extend([[[name, 0], possible_expression, [0, 0, 0]] + ['', ''] + ['', ''] + [str(e)]])
-            else:
-                continue
+            logger.error('Error in trying to process pandas expression: ' +  str(e))
+            continue
         except:
-            logging.Error('Error in trying to process pandas expression: UNKNOWN ERROR')
-            if solvency:
-                patterns.extend([[[name, 0], possible_expression, [0, 0, 0]] + ['',''] + ['', ''] + ['UNKNOWN ERROR']])
-            else:
-                continue
+            logger.error('Error in trying to process pandas expression: UNKNOWN ERROR')
+            continue
 
     return patterns
 
 def derive_patterns_from_code(metapattern = None,
                                 dataframe = None):
-    '''Derive conditional rule patterns
-       If no columns are given, then the algorithm searches for all possibilities
+    '''Derive patterns from code
+       Splits the patterns into conditional and quantitative
     '''
     patterns = list()
 
@@ -535,7 +585,7 @@ def derive_patterns_from_code(metapattern = None,
     parameters = metapattern.get("parameters", {})
 
 
-
+    # conditional pattern
     if pattern == '-->':
         possible_expressions = derive_conditional_pattern(metapattern = metapattern, dataframe = dataframe)
         for expression in possible_expressions:
@@ -560,10 +610,14 @@ def derive_patterns_from_code(metapattern = None,
 
 def derive_conditional_pattern(dataframe = None,
                                metapattern = None):
-    '''Here we derive the patterns from the metapattern definitions
-       by evaluating the pandas expressions of all potential patterns
     '''
-    logging.info(' Calling function derive_conditional_pattern')
+    Derive conditional rule patterns
+    If no columns are given, then the algorithm searches for all possibilities
+    '''
+
+    logger = logging.getLogger(__name__)
+
+    logger.info(' Calling function derive_conditional_pattern')
 
     # get items from metapattern definition
     parameters = metapattern.get("parameters", {})
@@ -571,8 +625,8 @@ def derive_conditional_pattern(dataframe = None,
     encode = metapattern.get(ENCODE, {})
     P_columns = metapattern.get("P_columns", list(dataframe.columns.values))
     Q_columns = metapattern.get("Q_columns", list(dataframe.columns.values))
-    P_values = metapattern.get("P_values", ['@']*len(P_columns)) # in case we do not have values
-    Q_values = metapattern.get("Q_values", ['@']*len(Q_columns))
+    P_values = metapattern.get("P_values", ['[@]']*len(P_columns)) # in case we do not have values
+    Q_values = metapattern.get("Q_values", ['[@]']*len(Q_columns))
 
     confidence, support = get_parameters(parameters)
 
@@ -588,20 +642,27 @@ def derive_conditional_pattern(dataframe = None,
 
         expressions = []
         df_features = df_features.drop_duplicates(P_columns + Q_columns)
+
+        # search for all possible values of P and Q
         for idx in range(len(df_features.index)):
             P_values = list(df_features[P_columns].values[idx])
             Q_values = list(df_features[Q_columns].values[idx])
             expression = generate_conditional_expression(P_columns, P_values, Q_columns, Q_values, parameters)
+            expression = expression.replace('"[@]"', '[@]')
 
             expressions.append(expression)
 
         return expressions
     # In the case that P and Q values are both given, we only want to compute it for these values and not search for other values like above
     expression = generate_conditional_expression(P_columns, P_values, Q_columns, Q_values, parameters)
+    expression = expression.replace('"[@]"', '[@]')
     return [expression]
 
 
 def get_parameters(parameters):
+    """
+    Gives confidence and support, also if not given in parameters
+    """
     confidence = parameters.get("min_confidence", 0.75)
     support = parameters.get("min_support", 2)
     return confidence, support
@@ -615,7 +676,14 @@ def derive_quantitative_pattern(metapattern = None,
                                 columns = None,
                                 value = None,
                                 parameters = {}):
-    logging.info(' Calling function derive_quantitative_pattern')
+
+    """
+    Derives quantitative pattern by splitting it into the right category
+    """
+
+    logger = logging.getLogger(__name__)
+
+    logger.info(' Calling function derive_quantitative_pattern')
 
     confidence, support = get_parameters(parameters)
     decimal = parameters.get("decimal", 8)
@@ -654,7 +722,8 @@ def derive_quantitative_pattern(metapattern = None,
 
     data_array = dataframe.values.T
     patterns = list()
-    if value is not None:
+
+    if value is not None: # If value is given, compare columns to value
         values = patterns_column_value(dataframe  = dataframe,
                                 value = value,
                                 pattern = pattern,
@@ -664,7 +733,7 @@ def derive_quantitative_pattern(metapattern = None,
         for val in values:
             patterns.extend(val)
 
-    elif pattern == 'percentile':
+    elif pattern == 'percentile': # If percentile, find that
         values = patterns_percentile(dataframe  = dataframe,
                                 pattern = pattern,
                                  pattern_name = pattern_name,
@@ -673,7 +742,7 @@ def derive_quantitative_pattern(metapattern = None,
         for val in values:
             patterns.extend(val)
 
-    elif pattern == 'sum':
+    elif pattern == 'sum': # If sum, then find a sum pattern
         sums = patterns_sums_column(dataframe  = dataframe,
                                  pattern_name = pattern_name,
                                  P_columns  = P_columns,
@@ -685,7 +754,8 @@ def derive_quantitative_pattern(metapattern = None,
     elif pattern == 'ratio':
         # TO DO
         return
-    else:
+
+    else: # Compare columns with eachother
         compares = patterns_column_column(dataframe  = dataframe,
                                 pattern = pattern,
                                  pattern_name = pattern_name,
@@ -724,14 +794,15 @@ def patterns_column_value(dataframe  = None,
                            pattern_name = "value",
                            columns = None,
                            parameters = {}):
-
+    '''Generate patterns of the form [c1] operator v1 where c1 is in columns and v1 is a value
+    '''
     confidence, support = get_parameters(parameters)
     data_array = dataframe.values.T
-
+    # Search for all the given columns
     for c in columns:
-        co = reduce(operators[pattern], [data_array[c, :], value])
+        co = reduce(operators[pattern], [data_array[c, :], value]) # Quick operations to check
         co_sum, ex_sum, conf = derive_pattern_statistics(co)
-
+        # If it fulfills the conditions
         if (conf >= confidence) and (co_sum >= support):
             possible_expression = generate_single_expression([dataframe.columns[c]], value, pattern)
             pandas_expressions = to_pandas_expressions(possible_expression, {}, parameters, dataframe)
@@ -744,25 +815,28 @@ def patterns_percentile(dataframe  = None,
                            pattern_name = "percentile",
                            columns = None,
                            parameters = {}):
-
+    '''Generate patterns of the form [c1] <= v1 and >= v2 where c1 is in columns, and v1 and v2 are the upper and lower percentile
+    '''
     confidence, support = get_parameters(parameters)
     data_array = dataframe.values.T
 
     percentile = parameters['percentile']
     add_per = (100-percentile)/2
     for c in columns:
-        # confirmations and exceptions of the pattern, a list of booleans
+        # Calculate upper and lower percentile
         upper = round(np.percentile(data_array[c, :], percentile + add_per),2)
         lower = round(np.percentile(data_array[c, :], add_per),2)
 
         co_up = reduce(operators['<='], [data_array[c, :], upper])
         co_down = reduce(operators['>='], [data_array[c, :], lower])
+
+        # Calculate which falls in the boundaries (list of bool)
         co = np.logical_and.reduce([co_up, co_down])
 
         co_sum, ex_sum, conf = derive_pattern_statistics(co)
 
         if (conf >= confidence) and (co_sum >= support):
-        # confirmations and exceptions of the pattern, a list of booleans
+            # Make expression
             possible_expression = generate_single_expression([dataframe.columns[c]], [lower, upper], pattern)
             pandas_expressions = to_pandas_expressions(possible_expression, {}, parameters, dataframe)
             xbrl_expressions = to_xbrl_expressions(possible_expression, {}, parameters)
@@ -778,6 +852,8 @@ def patterns_column_column(dataframe  = None,
                            parameters = {}):
     '''Generate patterns of the form [c1] operator [c2] where c1 and c2 are in columns
     '''
+    logger = logging.getLogger(__name__)
+
     confidence, support = get_parameters(parameters)
     decimal = parameters.get("decimal", 0)
     parameters['nonzero'] = True
@@ -787,19 +863,22 @@ def patterns_column_column(dataframe  = None,
     count = 0
 
     preprocess_operator = preprocess[pattern]
+
     if pattern == "=":
-        duplicates = {}
+        duplicates = {} # no duplicates
     for c0 in P_columns:
         for c1 in Q_columns:
             count += 1
             if count == 40:
-                logging.warning(' More than 40 possibilities!')
+                logger.warning(' More than 40 possibilities!')
             if c0 != c1:
                 # applying the filter
                 data_filter = reduce(preprocess_operator, [nonzero[c] for c in [c0, c1]])
                 if data_filter.any():
                     data_array = initial_data_array[:, data_filter]
                     if data_array.any():
+
+                        # keep track of duplicates and Calculate using decimal
                         if pattern == "=":
                             co = np.abs(data_array[c0, :] - data_array[c1, :]) < 1.5 * 10**(-decimal)
                             if c0 in duplicates:
@@ -812,10 +891,12 @@ def patterns_column_column(dataframe  = None,
                                         duplicates[c1] = [c0]
                             else:
                                 duplicates[c1] = [c0]
+                        # if not = pattern then normal reduce
                         else:
                             co = reduce(operators[pattern], data_array[[c0, c1], :])
                         co_sum, ex_sum, conf = derive_pattern_statistics(co)
                         if (conf >= confidence) and (co_sum >= support):
+                            # generate expression
                             possible_expression = generate_single_expression([dataframe.columns[c0]], [dataframe.columns[c1]], pattern)
                             pandas_expressions = to_pandas_expressions(possible_expression, {}, parameters, dataframe)
                             xbrl_expressions = to_xbrl_expressions(possible_expression, {}, parameters)
@@ -827,8 +908,11 @@ def patterns_sums_column( dataframe  = None,
                          P_columns  = None,
                          Q_columns  = None,
                          parameters = {}):
-    '''Generate patterns of the form sum [c1-list] = [c2] where c1-list is column list and c2 is column
     '''
+    Generate patterns of the form sum [c1-list] = [c2] where c1-list is column list and c2 is column
+    '''
+    logger = logging.getLogger(__name__)
+
     confidence, support = get_parameters(parameters)
     sum_elements = parameters.get("sum_elements", 2)
     decimal = parameters.get("decimal", 0)
@@ -837,13 +921,7 @@ def patterns_sums_column( dataframe  = None,
     # set up boolean masks for nonzero items per column
     nonzero = (dataframe.values != 0).T
     n = len(dataframe.columns)
-    # setup matrix to speed up proces (under development)
-    # matrix = np.ones(shape = (n, n), dtype = bool)
-    # for c in itertools.combinations(range(n), 2):
-    #     v = (data_array[c[1], :] <= data_array[c[0], :] + 1).any() # all is too strict
-    #     matrix[c[0], c[1]] = v
-    #     matrix[c[1], c[0]] = ~v
-    # np.fill_diagonal(matrix, False)
+
 
     count = 0
     neg_col = [1]*len(Q_columns)
@@ -857,11 +935,16 @@ def patterns_sums_column( dataframe  = None,
             except:
                 continue
             lhs_column_list = [col for col in P_columns if (col != rhs_column)]
+
+            # make combinations
             for lhs_columns in itertools.combinations(lhs_column_list, lhs_elements):
                 count += 1
-                if count == 50:
-                    logging.warning(' More than 50 possibilities!')
+                if count == 50: # log if we have a lot
+                    logger.warning(' More than 50 possibilities!')
+
                 all_columns = lhs_columns + (rhs_column,)
+
+                # reduce to get boolean list with all columns
                 data_filter = np.logical_and.reduce(nonzero[all_columns, :])
                 if data_filter.any():
                     data_array = start_array[:, data_filter]
@@ -875,6 +958,7 @@ def patterns_sums_column( dataframe  = None,
                         pattern_data = [[[pattern_name, 0], possible_expression, [co_sum, ex_sum, conf]] + pandas_expressions + xbrl_expressions + ['']]
                         yield pattern_data
 
+
 def derive_ratio_pattern(dataframe  = None,
                    pattern_name = None,
                    P_columns  = None,
@@ -917,14 +1001,16 @@ def derive_ratio_pattern(dataframe  = None,
 
 def to_pandas_expressions(pattern, encode, parameters, dataframe):
     """Derive pandas code from the pattern definition string both confirmation and exceptions"""
-    logging.debug(' Pattern in: ' + pattern)
+    logger = logging.getLogger(__name__)
+
+    logger.debug(' Pattern in: ' + pattern)
     # preprocessing step
     res = preprocess_pattern(pattern, parameters)
     # datapoints to pandas, i.e. {column} -> df[column]
     res, nonzero_col = datapoints2pandas(res, encode)
     # expression to pandas, i.e. IF X=x THEN Y=y -> df[df[X]=x & df[Y]=y] for confirmations
     co_str, ex_str = expression2pandas(res, nonzero_col, parameters)
-    logging.debug(' Pandas out: ' + co_str)
+    logger.debug(' Pandas out: ' + co_str)
 
     return [co_str, ex_str]
 
@@ -980,9 +1066,7 @@ def update_statistics(dataframe = None,
                 df_patterns.loc[idx, ERROR] = 'ERROR unknown'
 
             df_new_patterns = df_patterns
-        # deleting the levels of the index to the columns
-        for level in range(len(dataframe.index.names)):
-            del dataframe[dataframe.index.names[level]]
+
     return df_new_patterns
 
 def get_encodings():
@@ -1000,17 +1084,21 @@ def derive_results(dataframe = None,
     '''Results (patterns applied to data) are derived
        All info of the patterns is included in the results
     '''
-    logging.info(' Calling function derive_results')
+    logger = logging.getLogger(__name__)
+
+    logger.info('Analyze started ...')
+    logger.info('Shape of df_patterns: ' + str(df_patterns.shape))
 
     if (P_dataframe is not None) and (Q_dataframe is not None):
         try:
             dataframe = P_dataframe.join(Q_dataframe)
         except:
-            print("Join of P_dataframe and Q_dataframe failed, overlapping columns?")
+            logger.error("Join of P_dataframe and Q_dataframe failed, overlapping columns?")
             return []
     encodings = get_encodings()
 
     if (dataframe is not None) and (df_patterns is not None):
+
         df = dataframe.copy()
         results = list()
         for idx in df_patterns.index:
@@ -1022,8 +1110,12 @@ def derive_results(dataframe = None,
             try:
                 results_ex = eval(pandas_ex, encodings, {'df': df, 'MAX': np.maximum, 'MIN': np.minimum, 'SUM': np.sum}).index.values.tolist()
                 results_co = eval(pandas_co, encodings, {'df': df, 'MAX': np.maximum, 'MIN': np.minimum, 'SUM': np.sum}).index.values.tolist()
+
+                # Get the correct P and Q values that were given for each row
                 colq = get_value(df_patterns.loc[idx, "pattern_def"], 2, 1)
                 colp = get_value(df_patterns.loc[idx, "pattern_def"], 1, 1)
+
+                # Get it in the richt format, list if necessary
                 if isinstance(colq, list):
                     dataframe['combined_q']= dataframe[colq].values.tolist()
                     colq_old = colq
@@ -1036,12 +1128,14 @@ def derive_results(dataframe = None,
                     colp = dataframe.columns.get_loc(colp)
                 if colq != None:
                     colq = dataframe.columns.get_loc(colq)
+
+                #
                 for i in results_ex:
                     k = dataframe.index.get_loc(i)
 
                     if colp != None:
                         values_p = dataframe.iloc[k, colp]
-                        if isinstance(values_p, object):
+                        if isinstance(values_p, pd.Series): # If we have a pandas series we might have duplicate indices
                             if len(values_p) > 1:
                                 values_p = 'Duplicate indices!'
                             else:
@@ -1049,8 +1143,9 @@ def derive_results(dataframe = None,
                     else:
                         values_p = ""
                     if colq != None:
+
                         values_q = dataframe.iloc[k, colq]
-                        if isinstance(values_q, object):
+                        if isinstance(values_q, pd.Series): # If we have a pandas series we might have duplicate indices
                             if len(values_q) > 1:
                                 values_q = 'Duplicate indices!'
                             else:
@@ -1058,18 +1153,6 @@ def derive_results(dataframe = None,
                     else:
                         values_q = ""
 
-                    # Encode if necessary
-                    # if encode != {}:
-                    #     if not isinstance(values_p, list):
-                    #         values_p = [values_p]
-                    #     if not isinstance(values_q, list):
-                    #         values_q = [values_q]
-                    #     for c in range(len(colq_old)):
-                    #         if colq_old[c] in encode.keys():
-                    #             values_q[c] = eval(str(encode[colq_old[c]])+ "(s)", encodings, {'s': values_q[c]})
-                    #     for c in range(len(colq_old)):
-                    #         if colq_old[c] in encode.keys():
-                    #             values_q[c] = eval(str(encode[colq_old[c]])+ "(s)", encodings, {'s': values_q[c]})
 
                     results.append([False,
                                     df_patterns.loc[idx, "pattern_id"],
@@ -1081,11 +1164,12 @@ def derive_results(dataframe = None,
                                     df_patterns.loc[idx, "pattern_def"],
                                     values_p,
                                     values_q])
+
                 for i in results_co:
                     k = dataframe.index.get_loc(i)
                     if colp != None:
                         values_p = dataframe.iloc[k, colp]
-                        if isinstance(values_p, object):
+                        if isinstance(values_p, pd.Series):# If we have a pandas series we might have duplicate indices
                             if len(values_p) > 1:
                                 values_p = 'Duplicate indices!'
                             else:
@@ -1094,7 +1178,7 @@ def derive_results(dataframe = None,
                         values_p = ""
                     if colq != None:
                         values_q = dataframe.iloc[k, colq]
-                        if isinstance(values_q, object):
+                        if isinstance(values_q, pd.Series):# If we have a pandas series we might have duplicate indices
                             if len(values_q) > 1:
                                 values_q = 'Duplicate indices!'
                             else:
@@ -1111,6 +1195,7 @@ def derive_results(dataframe = None,
                                     df_patterns.loc[idx, "pattern_def"],
                                     values_p,
                                     values_q])
+                logger.info("Pattern " + df_patterns.loc[idx, "pattern_id"] + " correctly parsed (#co=" +str(len(results_co)) + ", #ex=" + str(len(results_ex))+")")
             except TypeError as e:
                 results.append([True,
                                 df_patterns.loc[idx, "pattern_id"],
@@ -1122,7 +1207,7 @@ def derive_results(dataframe = None,
                                 df_patterns.loc[idx, "pattern_def"],
                                 df_patterns.loc[idx, ERROR],
                                 'BUG'])
-                logging.error(' Error in analyze: ' + str(e))
+                logger.error('Error in analyze: ' + str(e))
             except:
                 results.append([True,
                                 df_patterns.loc[idx, "pattern_id"],
@@ -1134,7 +1219,7 @@ def derive_results(dataframe = None,
                                 df_patterns.loc[idx, "pattern_def"],
                                 df_patterns.loc[idx, ERROR],
                                 'BUG'])
-                logging.error(' Error in analyze: UNKOWN')
+                logger.error('Error in analyze: UNKOWN')
 
         df_results = pd.DataFrame(data = results, columns = RESULTS_COLUMNS)
         df_results.sort_values(by = ["index", "confidence", "support"], ascending = [True, False, False], inplace = True)
@@ -1144,9 +1229,11 @@ def derive_results(dataframe = None,
         except:
             df_results.index = df_results.index
 
+    for level in range(len(dataframe.index.names)):
+        del dataframe[dataframe.index.names[level]]
     df_results = ResultDataFrame(df_results)
-    logging.info(' Rows in analyze: ' + str(df_results.shape[0]))
-
+    logger.info('Shape of df_results: ' + str(df_results.shape))
+    logger.info('Analyze ended ...')
     return df_results
 
 
